#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import pytest


def test_all_tools_are_listed():
    # Note the ordering of this test (see: f1d744da16)
    from diffoscope.comparators import ComparatorManager
    from diffoscope.external_tools import EXTERNAL_TOOLS
    from diffoscope.tools import tool_required

    ComparatorManager().reload()

    tools = set(tool_required.all)

    for x in tools:
        if x not in EXTERNAL_TOOLS:
            pytest.fail(f"{x} is not present in EXTERNAL_TOOLS")


def test_get_tools():
    # Note the ordering of this test (see: f1d744da16)
    from diffoscope.comparators import ComparatorManager
    from diffoscope.tools import get_tools

    ComparatorManager().reload()

    tools = get_tools()
    missing_tools = get_tools(only_missing=True)
    k = "External-Tools-Required"
    for x in missing_tools[k]:
        if x not in tools[k]:
            pytest.fail(
                f"{x} must be present for {k} in tools and only_missing"
            )


def test_sbin_added_to_path():
    from diffoscope.tools import tool_required

    _, _, filenames = list(os.walk("/sbin"))[0]

    @tool_required(filenames[0])
    def fn():
        pass

    fn()


def test_required_tool_not_found():
    from diffoscope.exc import RequiredToolNotFound
    from diffoscope.tools import tool_required

    @tool_required("does-not-exist")
    def fn():
        pass

    with pytest.raises(RequiredToolNotFound):
        fn()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import logging
import plistlib
import subprocess

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command, our_check_output

logger = logging.getLogger(__name__)


class OpenSSLPKCS7(Command):
    @tool_required("openssl")
    def cmdline(self):
        return ("openssl", "pkcs7", "-print", "-noout", "-in", self.path)


class Pkcs7File(File):
    DESCRIPTION = "Public Key Cryptography Standards (PKCS) files (version #7)"
    FILE_TYPE_HEADER_PREFIX = b"-----BEGIN PKCS7-----"[:16]

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                OpenSSLPKCS7,
                self.path,
                other.path,
                source="openssl pkcs7 -print",
            )
        ]


class MobileProvisionFile(File):
    DESCRIPTION = "Apple Xcode mobile provisioning files"
    FILE_EXTENSION_SUFFIX = {".mobileprovision"}

    @staticmethod
    @tool_required("openssl")
    def _get_structured_profile_text(path):
        try:
            openssl_output = our_check_output(
                [
                    "openssl",
                    "smime",
                    "-inform",
                    "der",
                    "-verify",
                    "-noverify",
                    "-in",
                    path,
                ]
            )
        except subprocess.CalledProcessError as exc:
            logger.warning(
                "Failed to execute openssl to analayse Xcode file", exc
            )
            return "[Could not parse output]"

        return plistlib.dumps(plistlib.loads(openssl_output), sort_keys=True)

    def compare_details(self, other, source=None):
        my_content = MobileProvisionFile._get_structured_profile_text(
            self.path
        )
        other_content = MobileProvisionFile._get_structured_profile_text(
            other.path
        )
        return [
            Difference.from_text(
                my_content,
                other_content,
                self.name,
                other.name,
                source="openssl smime",
            )
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2018-2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import subprocess

from .utils.tools import skip_unless_tool_is_at_least


def black_version():
    try:
        out = subprocess.check_output(("black", "--version"))
    except subprocess.CalledProcessError as exc:
        out = exc.output

    # black --version format changed starting in 21.11b0. Returning the first
    # token starting with a decimal digit, since its ordinal position may vary.
    return [
        x for x in out.strip().decode("utf-8").split(" ") if x[0].isdigit()
    ]


@skip_unless_tool_is_at_least("black", black_version, "26.1.0")
def test_code_is_black_clean():
    output = subprocess.check_output(
        ("black", "--diff", "."), stderr=subprocess.PIPE
    ).decode("utf-8")

    # Display diff in "captured stdout call"
    if output:
        print(output)

    assert not output, output



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.config import Config
from diffoscope.comparators.missing_file import MissingFile
from diffoscope.comparators.gettext import MoFile

from ..utils.data import data, load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist

mo1 = load_fixture("test1.mo")
mo2 = load_fixture("test2.mo")


def test_identification(mo1):
    assert isinstance(mo1, MoFile)


def test_no_differences(mo1):
    difference = mo1.compare(mo1)
    assert difference is None


@pytest.fixture
def differences(mo1, mo2):
    return mo1.compare(mo2).details


@skip_unless_tools_exist("msgunfmt")
def test_diff(differences):
    assert_diff(differences[0], "mo_expected_diff")


mo_no_charset = load_fixture("test_no_charset.mo")
mo_iso8859_1 = load_fixture("test_iso8859-1.mo")


@skip_unless_tools_exist("msgunfmt")
def test_charsets(mo_no_charset, mo_iso8859_1):
    difference = mo_no_charset.compare(mo_iso8859_1)
    expected_diff = open(
        data("mo_charsets_expected_diff"), encoding="utf-8"
    ).read()
    assert difference.details[0].unified_diff == expected_diff


@skip_unless_tools_exist("msgunfmt")
def test_compare_non_existing(monkeypatch, mo1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = mo1.compare(MissingFile("/nonexisting", mo1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import shutil
import pytest

from diffoscope.config import Config
from diffoscope.comparators.gzip import GzipFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.missing_file import MissingFile
from diffoscope.comparators.utils.specialize import (
    specialize,
    is_direct_instance,
)

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_file_version_is_at_least

gzip1 = load_fixture("test1.gz")
gzip2 = load_fixture("test2.gz")
gzip3 = load_fixture("debian-bug-876316-control.tar.gz")


def test_identification(gzip1):
    assert is_direct_instance(gzip1, GzipFile)


def test_fallback_recognizes(gzip3):
    # the below always-True assertion is just to document the fact that we
    # should identify it correctly regardless of any bugs in file(1)
    assert (
        "gzip" not in gzip3.magic_file_type or "gzip" in gzip3.magic_file_type
    )
    assert is_direct_instance(gzip3, GzipFile)


def test_no_differences(gzip1):
    difference = gzip1.compare(gzip1)
    assert difference is None


@pytest.fixture
def differences(gzip1, gzip2):
    return gzip1.compare(gzip2).details


@skip_unless_file_version_is_at_least("5.37")
def test_metadata(differences):
    assert differences[0].source1.startswith("filetype")
    assert differences[0].source2.startswith("filetype")
    expected_diff = get_data("gzip_metadata_expected_diff")
    assert differences[0].unified_diff == expected_diff


def test_content_source(differences):
    assert differences[1].source1 == "test1"
    assert differences[1].source2 == "test2"


def test_content_source_without_extension(tmpdir, gzip1, gzip2):
    path1 = str(tmpdir.join("test1"))
    path2 = str(tmpdir.join("test2"))
    shutil.copy(gzip1.path, path1)
    shutil.copy(gzip2.path, path2)
    gzip1 = specialize(FilesystemFile(path1))
    gzip2 = specialize(FilesystemFile(path2))
    difference = gzip1.compare(gzip2).details
    assert difference[1].source1 == "test1-content"
    assert difference[1].source2 == "test2-content"


def test_content_diff(differences):
    expected_diff = get_data("text_ascii_expected_diff")
    assert differences[1].unified_diff == expected_diff


def test_compare_non_existing(monkeypatch, gzip1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = gzip1.compare(MissingFile("/nonexisting", gzip1))
    assert difference.source2 == "/nonexisting"
    assert difference.details[-1].source2 == "/dev/null"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017 Ximin Luo <infinity0@debian.org>
# Copyright © 2017, 2020-2022, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import pytest
import subprocess

from diffoscope.tools import os_is_gnu
from diffoscope.comparators.gzip import GzipFile
from diffoscope.comparators.rdata import RdbFile

from ..utils.data import load_fixture, get_data, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least

rdb1 = load_fixture("test1.rdb")
rdb2 = load_fixture("test2.rdb")
rdx1 = load_fixture("test1.rdx")
rdx2 = load_fixture("test2.rdx")


def rscript_version():
    val = subprocess.check_output(
        ("Rscript", "--version"), stderr=subprocess.STDOUT
    )

    m = re.search(r"version (?P<version>[^\(]+)\s", val.decode("utf-8"))
    if not m:
        return "~0unknown"

    return m.group("version")


def test_identification(rdb1, rdx1):
    assert isinstance(rdb1, RdbFile)
    assert isinstance(rdx1, GzipFile)


def test_no_differences(rdb1, rdx1):
    assert rdx1.compare(rdx1) is None
    assert rdx1.compare(rdx1) is None


@pytest.fixture
def differences_rdb(rdb1, rdb2):
    return rdb1.compare(rdb2).details


@pytest.fixture
def differences_rdx(rdx1, rdx2):
    return rdx1.compare(rdx2).details


@skip_unless_tools_exist("Rscript")
def test_num_items_rdb(differences_rdb):
    assert len(differences_rdb) == 1


@pytest.mark.skipif(
    not os_is_gnu(), reason="requires GNU stdlib for consistent %p formatting"
)
@skip_unless_tools_exist("Rscript")
@skip_unless_tool_is_at_least("Rscript", rscript_version, "4.2.0")
def test_item_rdb(differences_rdb):
    assert differences_rdb[0].source1.startswith("Rscript")
    assert_diff(differences_rdb[0], "rdb_expected_diff")


@skip_unless_tools_exist("Rscript")
def test_num_items_rdx(differences_rdx):
    assert len(differences_rdx) == 1


@skip_unless_tools_exist("Rscript")
def test_item_rdx(differences_rdx):
    assert differences_rdx[0].source1 == "test1.rdx-content"
    assert differences_rdx[0].source2 == "test2.rdx-content"
    assert_diff(differences_rdx[0].details[0], "rds_expected_diff")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2022, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import json
import os
import pprint
import logging
import collections

from diffoscope.difference import Difference
from diffoscope.tools import python_module_missing

from .utils.file import File

logger = logging.getLogger(__name__)

try:
    import jsondiff
except ImportError:  # noqa
    python_module_missing("jsondiff")
    jsondiff = None


class JSONFile(File):
    DESCRIPTION = "JSON files"
    FILE_TYPE_RE = re.compile(r"^JSON (?:text )?data")  # Requires file 5.35+

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        with open(file.path, "rb") as f:
            try:
                file.parsed = json.loads(
                    f.read().decode("utf-8", errors="ignore"),
                    object_pairs_hook=collections.OrderedDict,
                )
            except ValueError:
                return False

        return True

    def compare_details(self, other, source=None):
        difference = Difference.from_text(
            self.dumps(self),
            self.dumps(other),
            self.path,
            other.path,
            source="Pretty-printed",
        )

        if difference:
            self.compare_with_jsondiff(difference, other)

            return [difference]

        difference = Difference.from_text(
            self.dumps(self, sort_keys=False),
            self.dumps(other, sort_keys=False),
            self.path,
            other.path,
            source="Pretty-printed",
            comment="Ordering differences only",
        )

        return [difference]

    def compare_with_jsondiff(self, difference, other):
        if jsondiff is None:
            return

        # Ignore files over 100 KiB as jsondiff scales in O(n^2) time.
        # <https://github.com/xlwings/jsondiff/issues/73>
        limit = 100 * 1024
        if (
            os.stat(self.path).st_size > limit
            or os.stat(other.path).st_size > limit
        ):
            return

        logger.debug(f"Comparing using jsondiff {jsondiff.__version__}")

        a = getattr(self, "parsed", {})
        b = getattr(other, "parsed", {})

        try:
            diff = {repr(x): repr(y) for x, y in jsondiff.diff(a, b).items()}
        except Exception:
            return

        similarity = jsondiff.similarity(a, b)
        if similarity:
            difference.add_comment(f"Similarity: {similarity}%")

        differences = pprint.pformat(diff, width=100)
        if len(differences) > 512:
            differences = f"{differences[:512]} […]"
        difference.add_comment(f"Differences: {differences}")

    @staticmethod
    def dumps(file, sort_keys=True):
        if not hasattr(file, "parsed"):
            return ""
        return json.dumps(file.parsed, indent=4, sort_keys=sort_keys)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.gif import GifFile
from diffoscope.config import Config

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

gif1 = load_fixture("test1.gif")
gif2 = load_fixture("test2.gif")
gif3 = load_fixture("test3.gif")
gif4 = load_fixture("test4.gif")


def test_identification(gif1):
    assert isinstance(gif1, GifFile)


def test_no_differences(gif1):
    difference = gif1.compare(gif1)
    assert difference is None


@pytest.fixture
def differences(gif1, gif2):
    return gif1.compare(gif2).details


@skip_unless_tools_exist("gifbuild")
def test_diff(differences):
    assert_diff(differences[0], "gif_expected_diff")


@skip_unless_tools_exist("gifbuild")
def test_compare_non_existing(monkeypatch, gif1):
    assert_non_existing(monkeypatch, gif1, has_null_source=False)


@skip_unless_tools_exist("gifbuild", "convert", "identify")
def test_has_visuals(monkeypatch, gif3, gif4):
    monkeypatch.setattr(Config(), "compute_visual_diffs", True)
    gif_diff = gif3.compare(gif4)
    assert len(gif_diff.details) == 2
    assert len(gif_diff.details[1].visuals) == 2
    assert gif_diff.details[1].visuals[0].data_type == "image/png;base64"
    assert gif_diff.details[1].visuals[1].data_type == "image/gif;base64"


@skip_unless_tools_exist("gifbuild", "convert", "identify")
def test_no_visuals_different_size(monkeypatch, gif1, gif2):
    monkeypatch.setattr(Config(), "compute_visual_diffs", True)
    gif_diff = gif1.compare(gif2)
    assert len(gif_diff.details) == 1
    assert len(gif_diff.details[0].visuals) == 0



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2022, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import functools
import subprocess

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command, our_check_output


class Odt2txt(Command):
    @tool_required("odt2txt")
    def cmdline(self):
        # LibreOffice provides a "odt2txt" binary with different command-line
        # options.
        if self.odt2txt_variant() == "unoconv":
            return ("odt2txt", "--stdout", self.path)

        return ("odt2txt", "--width=-1", self.path)

    @staticmethod
    @functools.lru_cache
    def odt2txt_variant():
        try:
            out = our_check_output(["odt2txt", "--version"])
        except subprocess.CalledProcessError as e:
            out = e.output
        return out.decode("UTF-8").splitlines()[0].split()[0].strip()


class OdtFile(File):
    DESCRIPTION = "OpenOffice .odt files"
    FILE_TYPE_RE = re.compile(r"^OpenDocument Text\b")

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                Odt2txt, self.path, other.path, source="odt2txt"
            )
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2020, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import abc
import signal
import logging
import re
import subprocess

from .operation import Operation
from ...utils import format_cmdline

re_ansi_escapes = re.compile(rb"\x1b[^m]*m")

logger = logging.getLogger(__name__)


class Command(Operation, metaclass=abc.ABCMeta):
    MASK_STDERR = False
    MAX_STDERR_LINES = 25
    VALID_RETURNCODES = {0}

    def start(self):
        logger.debug("Executing %s", self.full_name())

        self._stdin = self.stdin()
        # "stdin" used to be a feeder but we didn't need the functionality so
        # it was simplified into the current form. it can be recovered from git
        # the extra functionality is needed in the future. alternatively,
        # consider using a shell pipeline ("sh -ec $script") to implement what
        # you need, because that involves much less code - like it or not (I
        # don't) shell is still the most readable option for composing processes
        self._process = subprocess.run(
            self.cmdline(),
            close_fds=True,
            env=self.env(),
            input=self.input(),
            stdin=self._stdin,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        self.error_string = self._read_stderr()
        self.returncode = self._process.returncode

    def stdin(self):
        return None

    @abc.abstractmethod
    def cmdline(self):
        raise NotImplementedError()

    @property
    def name(self):
        return self.cmdline()[0]

    def full_name(self, *args, **kwargs):
        kwargs.setdefault("replace", (self.path,))
        return format_cmdline(self.cmdline(), *args, **kwargs)

    def env(self):
        return None  # inherit parent environment by default

    def poll(self):
        pass

    def input(self):
        pass

    def _read_stderr(self):
        if self.MASK_STDERR:
            return ""

        buf = ""
        lines = self._process.stderr.splitlines(True)

        for index, line in enumerate(lines):
            if index >= Command.MAX_STDERR_LINES:
                break
            buf += line.decode("utf-8", errors="replace")

        if len(lines) > Command.MAX_STDERR_LINES:
            buf += "[ truncated after {} lines; {} ignored ]\n".format(
                Command.MAX_STDERR_LINES, len(lines) - Command.MAX_STDERR_LINES
            )

        return buf

    def should_show_error(self):
        # Handle SIGTERM once here as VALID_RETURNCODES may be overriden by
        # subclasses, and we don't want them to have to include it each time
        if self.returncode == -signal.SIGTERM:
            return False

        return self.returncode not in self.VALID_RETURNCODES

    @property
    def output(self):
        return self._process.stdout.splitlines(True)


def our_check_output(cmd, *args, **kwargs):
    logger.debug("Calling external command: %s", " ".join(cmd))

    return subprocess.check_output(cmd, *args, **kwargs)


def strip_ansi_escapes(val):
    return re_ansi_escapes.sub(b"", val)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020-2021, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.comparators.odt import OdtFile

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tool_is_at_least

from ..utils.nonexisting import assert_non_existing

odt1 = load_fixture("test1.odt")
odt2 = load_fixture("test2.odt")


def odt2txt_version():
    try:
        out = subprocess.check_output(["odt2txt", "--version"])
    except subprocess.CalledProcessError as e:
        out = e.output
    return out.decode("UTF-8").splitlines()[0].split()[1].strip()


def test_identification(odt1):
    assert isinstance(odt1, OdtFile)


def test_no_differences(odt1):
    difference = odt1.compare(odt1)
    assert difference is None


@pytest.fixture
def differences(odt1, odt2):
    return odt1.compare(odt2).details


@skip_unless_tool_is_at_least("odt2txt", odt2txt_version, "0.5")
def test_diff(differences):
    assert_diff(differences[0], "odt_expected_diff")


@skip_unless_tool_is_at_least("odt2txt", odt2txt_version, "0.5")
def test_compare_non_existing(monkeypatch, odt1):
    assert_non_existing(monkeypatch, odt1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2018-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import subprocess

from diffoscope.tools import tool_required
from diffoscope.tempfiles import get_named_temporary_file
from diffoscope.difference import Difference

from .utils.file import File

from .missing_file import MissingFile


class GnumericFile(File):
    DESCRIPTION = "Gnumeric spreadsheets"
    FILE_EXTENSION_SUFFIX = {".gnumeric"}

    @tool_required("ssconvert")
    def compare_details(self, other, source=None):
        if isinstance(other, MissingFile):
            return [
                Difference(
                    self.name,
                    other.name,
                    comment="Trying to compare two non-existing files.",
                )
            ]

        return [
            Difference.from_text(
                self.dump(self),
                self.dump(other),
                self.name,
                other.name,
                source="ssconvert",
            )
        ]

    def dump(self, file):
        t = get_named_temporary_file()

        subprocess.check_call(
            (
                "ssconvert",
                "--export-type=Gnumeric_stf:stf_assistant",
                file.path,
                t.name,
            )
        )

        with open(t.name) as f:
            return f.read().strip()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2016, 2018-2024 Chris Lamb <lamby@debian.org>
# Copyright © 2023 Mattia Rizzolo <mattia@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import re

from diffoscope.tools import (
    python_module_missing,
    tool_required,
    get_package_provider,
)
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command

logger = logging.getLogger(__name__)

PYPDF_MAJOR_VERSION = None

try:
    try:
        # PyPDF 3.x
        import pypdf
    except ImportError:
        # PyPDF2 1.x-2.x
        import PyPDF2 as pypdf

    try:
        try:
            # pyPDF 3.x
            from pypdf import PdfReader

            PYPDF_MAJOR_VERSION = 3
        except ImportError:
            # pyPDF 2.x
            import pypdf.PdfReader as PdfReader

            PYPDF_MAJOR_VERSION = 2
    except ImportError:
        # PyPDF2 1.x
        import pypdf.PdfFileReader as PdfReader

        PYPDF_MAJOR_VERSION = 1

    try:
        # PyPDF 3.x
        from pypdf.errors import PdfReadError
    except ImportError:
        try:
            # PyPDF 2.x
            from PyPDF2.errors import PdfReadError
        except ImportError:
            # PyPDF 1.x
            from PyPDF2.utils import PdfReadError

except (ImportError, RuntimeError):  # noqa
    python_module_missing("pypdf")
    pypdf = None


class Pdftotext(Command):
    @tool_required("pdftotext")
    def cmdline(self):
        return ["pdftotext", self.path, "-"]


class Dumppdf(Command):
    @tool_required("dumppdf")
    def cmdline(self):
        return ["dumppdf", "-at", self.path]


class PdfFile(File):
    DESCRIPTION = "PDF documents"
    FILE_TYPE_RE = re.compile(r"^PDF document\b")

    def compare_details(self, other, source=None):
        xs = []

        xs.append(Difference.from_operation(Pdftotext, self.path, other.path))

        # Include verbose dumppdf output unless we won't see any differences
        # without it.
        if not any(xs):
            xs.append(
                Difference.from_operation(Dumppdf, self.path, other.path)
            )

        # Always preppend metadata differences
        xs[:0] = self.gen_metadata_differences(other)

        return xs

    def gen_metadata_differences(self, other):
        if PYPDF_MAJOR_VERSION is None:
            pkg = get_package_provider("pypdf")
            infix = f" from the '{pkg}' package " if pkg else " "
            self.add_comment(
                f"Installing the 'pypdf' Python module{infix}"
                "may produce a better output."
            )

            return

        difference = Difference.from_text(
            self.dump_pypdf_metadata(self),
            self.dump_pypdf_metadata(other),
            self.name,
            other.name,
        )
        if difference:
            difference.add_comment("Document info")
        yield difference

        difference = Difference.from_text(
            self.dump_pypdf_annotations(self),
            self.dump_pypdf_annotations(other),
            self.name,
            other.name,
        )
        if difference:
            difference.add_comment("Annotations")
        yield difference

    def dump_pypdf_metadata(self, file):
        try:
            pdf = PdfReader(file.path)
            document_info = pdf.metadata

            if document_info is None:
                return ""

            xs = []
            for k, v in sorted(document_info.items()):
                xs.append("{}: {!r}".format(k.lstrip("/"), v))

            return "\n".join(xs)
        except PdfReadError as e:
            msg = "Could not extract pypdf metadata from {}: {}".format(
                os.path.basename(file.name), e
            )
            self.add_comment(msg)
            logger.error(msg)
            return ""

    def dump_pypdf_annotations(self, file):
        try:
            pdf = PdfReader(file.path)

            xs = []
            for x in range(len(pdf.pages)):
                page = pdf.pages[x]

                try:
                    for annot in page["/Annots"]:
                        if PYPDF_MAJOR_VERSION == 3:
                            subtype = annot.get_object()["/Subtype"]
                            if subtype == "/Text":
                                xs.append(annot.get_object()["/Contents"])
                        else:
                            subtype = annot.getObject()["/Subtype"]
                            if subtype == "/Text":
                                xs.append(annot.getObject()["/Contents"])
                except KeyError:
                    pass

            return "\n".join(xs)
        except PdfReadError as e:
            msg = "Could not extract pypdf annotations from {}: {}".format(
                os.path.basename(file.name), e
            )
            file.add_comment(msg)
            logger.error(msg)
            return ""



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2018-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import html
import re

re_diff_line_numbers = re.compile(
    r"(^|\n)@@ -(\d+),(\d+) \+(\d+),(\d+) @@(?=\n|$)"
)


def diff_ignore_line_numbers(diff):
    return re_diff_line_numbers.sub(r"\1@@ -XX,XX +XX,XX @@", diff)


def _collapse_line(line, escape=html.escape):
    len_ = len(escape(line))
    return str(len_ - 1) + "\n" if line[-1] == "\n" else str(len_)


def _diff_collapse_line(line):
    return (
        line[0] + _collapse_line(line[1:])
        if line and line[0] in "+- "
        else line
    )


def _expand_line(line):
    return (
        (int(line[:-1]) * ".") + "\n"
        if line[-1] == "\n"
        else (int(line) * ".")
    )


def _diff_expand_line(line):
    return (
        line[0] + _expand_line(line[1:]) if line and line[0] in "+- " else line
    )


def diff_collapse(diff):
    return diff.map_lines(_diff_collapse_line, _collapse_line)


def diff_expand(diff):
    return diff.map_lines(_diff_expand_line, _expand_line)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2023 FC Stegerman <flx@obfusk.net>
# Copyright © 2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.config import Config
from diffoscope.comparators.arsc import ArscFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist

arsc1 = load_fixture("resources1.arsc")
arsc2 = load_fixture("resources2.arsc")


def test_identification(arsc1):
    assert isinstance(arsc1, ArscFile)


def test_no_differences(arsc1):
    difference = arsc1.compare(arsc1)
    assert difference is None


@pytest.fixture
def differences(arsc1, arsc2):
    return arsc1.compare(arsc2).details


def check_arsc_differences(differences, expected_diff):
    assert differences[0].source1 == "aapt2 dump resources {}"
    assert differences[0].source2 == "aapt2 dump resources {}"
    assert expected_diff == differences[0].unified_diff


@skip_unless_tools_exist("aapt2")
@pytest.mark.xfail(strict=False)
def test_differences(differences):
    expected_diff = get_data("arsc_expected_diff")
    check_arsc_differences(differences, expected_diff)


@skip_unless_tools_exist("aapt2")
def test_compare_non_existing(monkeypatch, arsc1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = arsc1.compare(MissingFile("/nonexisting", arsc1))
    assert difference.source2 == "/nonexisting"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017 Ximin Luo <infinity0@debian.org>
# Copyright © 2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.main import main
from diffoscope.comparators.utils.compare import compare_root_paths
from diffoscope.readers import load_diff_from_path

from .utils.data import cwd_data, get_data


def run_read_write(capsys, diff, *args):
    with pytest.raises(SystemExit) as exc, cwd_data():
        main(args + (f"--load-existing-diff={diff}",))

    out, err = capsys.readouterr()

    assert err == ""
    assert exc.value.code == 1
    assert out == get_data(diff)  # presented-output is same as parsed-input
    return out


def run_diff_read(diffpath):
    with cwd_data():
        diff = compare_root_paths("test1.tar", "test2.tar")
        read = load_diff_from_path(diffpath)
    assert diff.equals(read)


def test_json(capsys):
    run_read_write(capsys, "output.json", "--json", "-")
    run_diff_read("output.json")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import io
import re
import logging

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command

logger = logging.getLogger(__name__)


class Msgunfmt(Command):
    CHARSET_RE = re.compile(rb'^"Content-Type: [^;]+; charset=([^\\]+)\\n"$')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._header = io.BytesIO()
        self._encoding = None

    @tool_required("msgunfmt")
    def cmdline(self):
        return ["msgunfmt", self.path]

    def filter(self, line):
        if not self._encoding:
            self._header.write(line)
            if line == b"\n":
                logger.debug(
                    "unable to determine PO encoding, let's hope it's utf-8"
                )
                self._encoding = "utf-8"
                return self._header.getvalue()
            found = Msgunfmt.CHARSET_RE.match(line)
            if found:
                self._encoding = found.group(1).decode("us-ascii").lower()
                return (
                    self._header.getvalue()
                    .decode(self._encoding)
                    .encode("utf-8")
                )
            return b""

        if self._encoding == "utf-8":
            return line

        return line.decode(self._encoding).encode("utf-8")


class MoFile(File):
    DESCRIPTION = "Gettext message catalogues"
    FILE_TYPE_RE = re.compile(r"^GNU message catalog\b")

    def compare_details(self, other, source=None):
        return [Difference.from_operation(Msgunfmt, self.path, other.path)]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2019-2020, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import pytest
import subprocess

from diffoscope.comparators.pgp import PgpFile, PgpSignature

from ..utils.data import load_fixture, assert_diff, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.versions import Version
from ..utils.nonexisting import assert_non_existing

pgp1 = load_fixture("test1.pgp")
pgp2 = load_fixture("test2.pgp")
signed1 = load_fixture("test1_signed.pgp")
signed2 = load_fixture("test2_signed.pgp")
signature1 = load_fixture("test1.asc")
signature2 = load_fixture("test2.asc")


def test_identification(pgp1):
    assert isinstance(pgp1, PgpFile)


def test_no_differences(pgp1):
    difference = pgp1.compare(pgp1)
    assert difference is None


def pgpdump_version():
    out = subprocess.check_output(
        ("pgpdump", "-v"), stderr=subprocess.STDOUT
    ).decode("utf-8")
    # pgpdump version 0.33, Copyright (C) 1998-2017 Kazu Yamamoto
    m = re.search(r"^pgpdump version (?P<version>\d+\.\d+)", out)
    if m is None:
        raise ValueError(f"Error parsing `pgpdump -v` output: {out}")

    return Version(m.group("version"))


@pytest.fixture
def differences(pgp1, pgp2):
    return pgp1.compare(pgp2).details


@skip_unless_tools_exist("pgpdump")
def test_diff(differences):
    assert_diff(differences[0], "pgp_expected_diff")


@skip_unless_tools_exist("pgpdump")
def test_compare_non_existing(monkeypatch, pgp1):
    assert_non_existing(monkeypatch, pgp1, has_null_source=False)


def test_pgp_signature_identification(signature1, signature2):
    assert isinstance(signature1, PgpSignature)
    assert isinstance(signature2, PgpSignature)


@skip_unless_tools_exist("pgpdump")
@skip_unless_tools_exist("gpg")
def test_pgp_signature(signature1, signature2):
    difference = signature1.compare(signature2)
    assert_diff(difference, "pgp_signature_expected_diff")
    assert len(difference.details) == 2
    assert difference.details[0].source1 == "pgpdump"
    assert difference.details[1].source1 == "gpg --list-packets"


@skip_unless_tools_exist("pgpdump")
def test_signed_identification(signed1):
    assert isinstance(signed1, PgpFile)


@pytest.fixture
def signed_differences(signed1, signed2):
    return signed1.compare(signed2).details


@skip_unless_tools_exist("pgpdump")
def test_signed_diff(signed_differences):
    expected_diff = get_data("pgp_signed_expected_diff")
    if pgpdump_version() > Version(0.33):
        expected_diff = expected_diff.replace(
            "Format - binary", "Packet data format - binary"
        )
        expected_diff = expected_diff.replace(
            "File modified time", "Creation time"
        )
    assert signed_differences[0].unified_diff == expected_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2018, 2020-2021, 2023-2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.pdf import PdfFile

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import (
    skip_unless_tools_exist,
    skip_unless_module_exists,
    skipif,
)
from ..utils.nonexisting import assert_non_existing

pdf1 = load_fixture("test1.pdf")
pdf2 = load_fixture("test2.pdf")
pdf3 = load_fixture("test3.pdf")
pdf4 = load_fixture("test4.pdf")
pdf1a = load_fixture("test_weird_non_unicode_chars1.pdf")
pdf2a = load_fixture("test_weird_non_unicode_chars2.pdf")


def skip_unless_pypdf3():
    def fn():
        try:
            import pypdf
        except (ImportError, RuntimeError):
            return True

        return not int(pypdf.__version__.split(".")[0]) >= 3

    return skipif(fn(), reason="pypdf not installed or not version 3.x+")


def test_identification(pdf1):
    assert isinstance(pdf1, PdfFile)


def test_no_differences(pdf1):
    difference = pdf1.compare(pdf1)
    assert difference is None


def test_differences_found_with_weird_encoding(pdf1a, pdf2a):
    # diffoscope used to crash here due to weird encoding
    difference = pdf1a.compare(pdf2a)
    assert difference


@pytest.fixture
def differences(pdf1, pdf2):
    return pdf1.compare(pdf2).details


@skip_unless_tools_exist("pdftotext")
def test_text_diff(differences):
    assert_diff(differences[0], "pdf_text_expected_diff")


@skip_unless_tools_exist("pdftotext")
def test_compare_non_existing(monkeypatch, pdf1):
    assert_non_existing(monkeypatch, pdf1, has_null_source=False)


@pytest.fixture
def differences_metadata(pdf1, pdf1a):
    return pdf1.compare(pdf1a).details


@skip_unless_pypdf3()
@skip_unless_tools_exist("pdftotext")
def test_metadata(differences_metadata):
    assert len(differences_metadata) == 2

    assert_diff(differences_metadata[0], "pdf_metadata_expected_diff")


@pytest.fixture
def differences_annotations(pdf3, pdf4):
    return pdf3.compare(pdf4).details


@skip_unless_pypdf3()
@skip_unless_tools_exist("pdftotext")
def test_annotations(differences_annotations):
    assert_diff(differences_annotations[0], "pdf_annotations_expected_diff")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2019-2020, 2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.pcap import PcapFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

pcap1 = load_fixture("test1.pcap")
pcap2 = load_fixture("test2.pcap")


def test_identification(pcap1):
    assert isinstance(pcap1, PcapFile)


def test_no_differences(pcap1):
    difference = pcap1.compare(pcap1)
    assert difference is None


@pytest.fixture
def differences(pcap1, pcap2):
    return pcap1.compare(pcap2).details


@skip_unless_tools_exist("tcpdump")
def test_diff(differences):
    expected_diff = get_data("pcap_expected_diff")
    assert differences[0].unified_diff[: 2**13] == expected_diff[: 2**13]


@skip_unless_tools_exist("tcpdump")
def test_compare_non_existing(monkeypatch, pcap1):
    assert_non_existing(monkeypatch, pcap1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2020, 2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.config import Config
from diffoscope.comparators.dex import DexFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import (
    skip_unless_tools_exist,
    skip_unless_tool_is_at_least,
    skip_unless_tool_is_between,
)

from .test_java import javap_version

dex1 = load_fixture("test1.dex")
dex2 = load_fixture("test2.dex")


def enjarify_version():
    # Module enjarify.typeinference appeared in enjarify 1.0.3.  We use a call
    # directly to the python3 binary over importing with this module to escape
    # virtualenvs and to match the behaviour of enjarify(1).
    if (
        subprocess.call(
            ("python3", "-c", "import enjarify.typeinference"),
            stderr=subprocess.PIPE,
        )
        == 0
    ):
        return "1.0.3"
    return "1.0.2"


def test_identification(dex1):
    assert isinstance(dex1, DexFile)


def test_no_differences(dex1):
    difference = dex1.compare(dex1)
    assert difference is None


@pytest.fixture
def differences(dex1, dex2):
    return dex1.compare(dex2).details


def check_dex_differences(differences, expected_diff):
    assert differences[0].source1 == "dexdump -a -d -f -h {}"
    assert differences[0].source2 == "dexdump -a -d -f -h {}"
    assert differences[1].source1 == "test1.jar"
    assert differences[1].source2 == "test2.jar"
    zipinfo = differences[1].details[0]
    classdiff = differences[1].details[1]
    assert zipinfo.source1 == "zipinfo -v {}"
    assert zipinfo.source2 == "zipinfo -v {}"
    assert classdiff.source1 == "com/example/MainActivity.class"
    assert classdiff.source2 == "com/example/MainActivity.class"
    found_diff = (
        differences[0].unified_diff
        + zipinfo.unified_diff
        + classdiff.details[0].unified_diff
    )
    assert expected_diff == found_diff


@skip_unless_tools_exist("enjarify", "zipinfo", "javap", "procyon", "dexdump")
@skip_unless_tool_is_between("javap", javap_version, "9.0.4", "14.0")
@skip_unless_tool_is_at_least("enjarify", enjarify_version, "1.0.3")
def test_differences(differences):
    expected_diff = get_data("dex_expected_diffs")
    check_dex_differences(differences, expected_diff)


@skip_unless_tools_exist("enjarify", "zipinfo", "javap", "dexdump", "procyon")
@skip_unless_tool_is_at_least("javap", javap_version, "14.0")
@skip_unless_tool_is_at_least("enjarify", enjarify_version, "1.0.3")
@pytest.mark.xfail(strict=False)
def test_javap_14_differences(differences):
    expected_diff = get_data("dex_javap_14_expected_diffs")
    check_dex_differences(differences, expected_diff)


@skip_unless_tools_exist("enjarify", "zipinfo", "javap")
@pytest.mark.xfail(strict=False)
def test_compare_non_existing(monkeypatch, dex1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = dex1.compare(MissingFile("/nonexisting", dex1))
    assert difference.source2 == "/nonexisting"
    assert difference.details[-1].source2 == "/dev/null"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2018-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.comparators.ocaml import OcamlInterfaceFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import assert_diff_startswith
from ..utils.tools import skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing


def ocaml_fixture(prefix):
    @pytest.fixture
    def cmi(tmpdir):
        input_ = str(tmpdir.join("{}.mli".format(prefix)))
        output = str(tmpdir.join("{}.cmi".format(prefix)))

        with open(input_, "w"):
            pass

        subprocess.check_call(("ocamlc", "-c", input_))

        return specialize(FilesystemFile(output))

    return cmi


cmi1 = ocaml_fixture("test1")
cmi2 = ocaml_fixture("test2")


def ocaml_version():
    try:
        out = subprocess.check_output(["ocaml", "-version"])
    except subprocess.CalledProcessError as e:
        out = e.output
    return out.decode("utf-8").split()[-1]


@skip_unless_tool_is_at_least("ocamlobjinfo", ocaml_version, "4.11")
def test_identification(cmi1):
    assert isinstance(cmi1, OcamlInterfaceFile)


@pytest.fixture
def differences(cmi1, cmi2):
    return cmi1.compare(cmi2).details


@skip_unless_tool_is_at_least("ocamlobjinfo", ocaml_version, "4.11")
def test_no_differences(cmi1):
    difference = cmi1.compare(cmi1)
    assert difference is None


@skip_unless_tool_is_at_least("ocamlobjinfo", ocaml_version, "4.11")
def test_diff(differences):
    assert_diff_startswith(differences[0], "ocaml_expected_diff")


@skip_unless_tool_is_at_least("ocamlobjinfo", ocaml_version, "4.11")
def test_compare_non_existing(monkeypatch, cmi1):
    assert_non_existing(monkeypatch, cmi1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2019-2020, 2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

from .utils import Presenter


class MarkdownTextPresenter(Presenter):
    def __init__(self, print_func):
        self.print_func = print_func
        super().__init__()

    def visit_difference(self, difference):
        if difference.source1 == difference.source2:
            self.title(difference.source1)
        else:
            self.title(
                "Comparing `{}` & `{}`".format(
                    difference.source1, difference.source2
                )
            )

        for x in difference.comments:
            self.print_func(f" * *{x}*")
            self.print_func()

        if difference.unified_diff:
            self.print_func("```diff")
            self.print_func(difference.unified_diff.rstrip())
            self.print_func("```")
            self.print_func()

    def title(self, val):
        prefix = "#" * min(self.depth + 1, 6)

        self.print_func("{} {}".format(prefix, val))
        self.print_func()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015      Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017 Mattia Rizzolo <mattia@debian.org>
# Copyright © 2017-2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import pytest
import functools
import importlib.util
import subprocess

from diffoscope.tools import get_package_provider, find_executable

from ..utils.versions import Version


def file_version():
    return (
        subprocess.check_output(("file", "-v"))
        .decode("utf-8")
        .splitlines()[0]
        .split("-")[-1]
    )


def file_version_is_lt(version):
    return Version(file_version()) < Version(version)


def file_version_is_ge(version):
    return Version(file_version()) >= Version(version)


def tools_missing(*required):
    return not required or any(find_executable(x) is None for x in required)


def skipif(*args, **kwargs):
    """
    Call `pytest.mark.skipif` with the specified arguments.

    If `check_env_for_missing_tools=True` is passed and the
    `DIFFOSCOPE_TESTS_FAIL_ON_MISSING_TOOLS` environment variable is exported,
    this alters the behaviour such that a tool listed within this variable is
    treated as a failed test instead of being skipped.

    For more information on the rationale here, see issue #35.
    """

    if not kwargs.get("check_env_for_missing_tools", False):
        return pytest.mark.skipif(*args, **kwargs)

    key = "DIFFOSCOPE_FAIL_TESTS_ON_MISSING_TOOLS"
    val = os.environ.get(key)

    if val is None:
        return pytest.mark.skipif(*args, **kwargs)

    tools_required = kwargs.get("tools", ())
    missing_tools = val.split() + ["/missing"]  # special value used in tests

    if not tools_required or any(
        x for x in tools_required if x in missing_tools
    ):
        return pytest.mark.skipif(*args, **kwargs)

    msg = "{} ({}={!r})".format(kwargs["reason"], key, val)

    # We cannot simply call pytest.fail here as that would result in a failure
    # during the test collection phase instead when the test is actually
    # executed.
    def outer(*args1, **kwargs1):
        def inner(*args2, **kwargs2):
            if args[0]:  # i.e. the condition of the skipif() is True
                return pytest.fail(msg)
            return None

        return inner

    return outer


def skip_unless_tools_exist(*required):
    return skipif(
        tools_missing(*required),
        reason=reason(*required),
        tools=required,
        check_env_for_missing_tools=True,
    )


def skip_if_tool_version_is(tool, actual_ver, target_ver, vcls=Version):
    if tools_missing(tool):
        return skipif(True, reason=reason(tool), tools=(tool,))
    if callable(actual_ver):
        actual_ver = actual_ver()
    return skipif(
        vcls(str(actual_ver)) == vcls(str(target_ver)),
        reason="requires {} != {}; {} detected".format(
            tool, target_ver, actual_ver
        ),
        tools=(tool,),
    )


def skip_unless_tool_is_at_least(tool, actual_ver, min_ver, vcls=Version):
    if tools_missing(tool) and module_is_not_importable(tool):
        return skipif(True, reason=reason(tool), tools=(tool,))
    if callable(actual_ver):
        actual_ver = actual_ver()
    return skipif(
        vcls(str(actual_ver)) < vcls(str(min_ver)),
        reason="{} >= {}; {} detected".format(
            reason(tool, force_include=True), min_ver, actual_ver
        ),
        tools=(tool,),
    )


def skip_unless_tool_is_at_most(tool, actual_ver, max_ver, vcls=Version):
    if tools_missing(tool) and module_is_not_importable(tool):
        return skipif(True, reason=reason(tool), tools=(tool,))
    if callable(actual_ver):
        actual_ver = actual_ver()
    return skipif(
        vcls(str(actual_ver)) > vcls(str(max_ver)),
        reason="{} <= {}; {} detected".format(
            reason(tool, force_include=True), max_ver, actual_ver
        ),
        tools=(tool,),
    )


def skip_unless_tool_is_between(
    tool, actual_ver, min_ver, max_ver, vcls=Version
):
    if tools_missing(tool):
        return skipif(True, reason=reason(tool), tools=(tool,))
    if callable(actual_ver):
        actual_ver = actual_ver()
    return skipif(
        (vcls(str(actual_ver)) < vcls(str(min_ver)))
        or (vcls(str(actual_ver)) > vcls(str(max_ver))),
        reason="{} min {} >= {}; {} detected".format(
            reason(tool, force_include=True), min_ver, max_ver, actual_ver
        ),
        tools=(tool,),
    )


def skip_if_binutils_does_not_support_x86():
    if tools_missing("objdump"):
        return skip_unless_tools_exist("objdump")

    return skipif(
        "elf64-x86-64" not in get_supported_elf_formats(),
        reason="requires a binutils capable of reading x86-64 binaries",
        tools=("objdump",),
    )


@functools.lru_cache()
def get_supported_elf_formats():
    try:
        return set(
            subprocess.check_output(("objdump", "--info"))
            .decode("utf-8")
            .splitlines()
        )
    except (subprocess.CalledProcessError, FileNotFoundError):
        return set()


def module_is_not_importable(x):
    try:
        if importlib.util.find_spec(x) is None:
            return True
        # an existent module is not necessarily importable, e.g. if its child
        # modules are not available, e.g. if we are running diffoscope using a
        # non-default version of python, and the module uses extension modules
        # that haven't been compiled for this version
        importlib.import_module(x)
    except ImportError:
        # Probing for submodules (eg. ``debian.deb822``) will attempt to
        # import ``debian`` so we must handle that failing.
        return True

    return False


def skip_unless_module_exists(name):
    return skipif(
        module_is_not_importable(name),
        reason=f"requires {name} Python module",
        tools=(f"{name}_module",),
    )


def skip_unless_file_version_is_at_least(version):
    return skip_unless_tool_is_at_least("file", file_version, version)


def reason(*tools, force_include=False):
    xs = []

    for x in tools:
        if not tools_missing(x) and not force_include:
            continue
        provider = get_package_provider(x)
        if provider is None:
            xs.append(x)
            continue

        xs.append("{}; try installing {}".format(x, provider))

    if not xs:
        return ""

    return "requires {}".format(" and ".join(xs))



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import subprocess

import pytest

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.openssl import Pkcs7File
from diffoscope.comparators.utils.specialize import specialize

from ..utils.tools import skip_unless_tools_exist


def pkcs7_fixture(prefix):
    @pytest.fixture
    def inner(tmpdir):
        crt = str(tmpdir.join("{}.crt".format(prefix)))
        pem = str(tmpdir.join("{}.pem".format(prefix)))
        p7c = str(tmpdir.join("{}.p7c".format(prefix)))

        subprocess.check_call(
            (
                "openssl",
                "req",
                "-x509",
                "-newkey",
                "rsa:4096",
                "-keyout",
                pem,
                "-out",
                crt,
                "-days",
                "365",
                "-nodes",
                "-subj",
                "/CN=localhost",
            )
        )
        subprocess.check_call(
            (
                "openssl",
                "crl2pkcs7",
                "-nocrl",
                "-certfile",
                crt,
                "-out",
                p7c,
                "-certfile",
                pem,
            )
        )

        return specialize(FilesystemFile(p7c))

    return inner


pkcs71 = pkcs7_fixture("test1")
pkcs72 = pkcs7_fixture("test2")


@skip_unless_tools_exist("openssl")
def test_identification(pkcs71):
    assert isinstance(pkcs71, Pkcs7File)


@skip_unless_tools_exist("openssl")
def test_no_differences(pkcs71):
    difference = pkcs71.compare(pkcs71)
    assert difference is None


@pytest.fixture
def differences(pkcs71, pkcs72):
    return pkcs71.compare(pkcs72).details


@skip_unless_tools_exist("openssl")
def test_differences(differences):
    # Don't test exact unified diff; the signatures generated in
    # `pkcs7_fixture` are non-deterministic.

    assert "notAfter:" in differences[0].unified_diff
    assert "serialNumber:" in differences[0].unified_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017 Juliana Oliveira R <juliana.orod@gmail.org>
# Copyright © 2017-2020, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

import itertools

from ..utils.data import load_fixture, get_data
from ..utils.tools import (
    skip_unless_tools_exist,
    file_version_is_ge,
    skip_unless_file_version_is_at_least,
)

gzip1 = load_fixture("containers/a.tar.gz")
gzip2 = load_fixture("containers/b.tar.gz")

xz1 = load_fixture("containers/a.tar.xz")
xz2 = load_fixture("containers/b.tar.xz")

bzip1 = load_fixture("containers/a.tar.bz2")
bzip2 = load_fixture("containers/b.tar.bz2")

TYPES = "gzip bzip2 xz".split()


@pytest.fixture
def set1(gzip1, bzip1, xz1):
    return dict(zip(TYPES, [gzip1, bzip1, xz1]))


@pytest.fixture
def set2(gzip2, bzip2, xz2):
    return dict(zip(TYPES, [gzip2, bzip2, xz2]))


def expected_magic_diff(ext1, ext2):
    magic = {
        "bzip2": "bzip2 compressed data, block size = 900k\n",
        "gzip": "gzip compressed data, last modified: Sun Sep 10 22:19:44 2017, from Unix\n",
        "xz": "XZ compressed data",
    }
    if "xz" in (ext1, ext2):
        if file_version_is_ge("5.40"):
            magic["xz"] += ", checksum CRC64\n"
        else:
            magic["xz"] += "\n"

    return "@@ -1 +1 @@\n" + "-" + magic[ext1] + "+" + magic[ext2]


def expected_type_diff(ext1, ext2):
    return "@@ -1 +1 @@\n-%sFile\n+%sFile\n" % (
        ext1.capitalize(),
        ext2.capitalize(),
    )


# Compares same content files, but with different extensions


@skip_unless_tools_exist("xz")
@skip_unless_file_version_is_at_least("5.37")
def test_equal(set1):
    for x, y in itertools.product(TYPES, TYPES):
        diff = set1[x].compare(set1[y])
        if x == y:
            assert diff is None
        else:
            differences = diff.details
            assert differences[0].unified_diff == expected_magic_diff(
                x, y
            ), "{} {}".format(x, y)
            assert differences[1].unified_diff == expected_type_diff(x, y)


# Compares different content files with different extensions


@skip_unless_tools_exist("xz")
@skip_unless_file_version_is_at_least("5.37")
def test_different(set1, set2):
    for x, y in itertools.product(TYPES, TYPES):
        expected_diff = get_data("containers/different_files_expected_diff")
        differences = set1[x].compare(set2[y]).details
        if x == y:
            assert differences[0].details[1].unified_diff == expected_diff
        else:
            assert differences[0].unified_diff == expected_magic_diff(
                x, y
            ), "{} {}".format(x, y)
            assert differences[1].unified_diff == expected_type_diff(x, y)
            assert differences[2].details[1].unified_diff == expected_diff



#!/usr/bin/env python3

import json

# Load extras_require dict from external JSON file. This allows it to be easily
# shared by the main setup.py script.
with open("extras_require.json") as f:
    extras_require = json.load(f)

xs = set(f"python3-{x}" for reqs in extras_require.values() for x in reqs)

print(", ".join(sorted(xs)))



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.hdf import Hdf5File
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_module_exists
from ..utils.nonexisting import assert_non_existing

hdf5_1 = load_fixture("test1.hdf5")
hdf5_2 = load_fixture("test2.hdf5")


def hdf5_fixture(prefix):
    @pytest.fixture
    def hdf5d(tmpdir):
        filename = str(tmpdir.join("{}.db".format(prefix)))

        # Listed in debian/tests/control.in
        import h5py

        with h5py.File(filename, "w") as f:
            f.create_dataset(prefix, (100,), dtype="i")

        return specialize(FilesystemFile(filename))

    return hdf5d


hdf5_1 = hdf5_fixture("test1")
hdf5_2 = hdf5_fixture("test2")


@skip_unless_module_exists("h5py")
def test_identification(hdf5_1):
    assert isinstance(hdf5_1, Hdf5File)


@skip_unless_module_exists("h5py")
def test_no_differences(hdf5_1):
    difference = hdf5_1.compare(hdf5_1)
    assert difference is None


@pytest.fixture
def differences(hdf5_1, hdf5_2):
    return hdf5_1.compare(hdf5_2).details


@skip_unless_tools_exist("h5dump")
@skip_unless_module_exists("h5py")
def test_diff(differences):
    assert_diff(differences[0], "hdf5_expected_diff")


@skip_unless_tools_exist("h5dump")
@skip_unless_module_exists("h5py")
def test_compare_non_existing(monkeypatch, hdf5_1):
    assert_non_existing(monkeypatch, hdf5_1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2021 Brent Spillner <s p i l l n e r @ a c m . o r g>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import stat
import logging

from diffoscope.tempfiles import get_named_temporary_file
from diffoscope.difference import Difference

from .binary import FilesystemFile
from .utils.file import File

logger = logging.getLogger(__name__)


class SocketOrFIFO(File):
    DESCRIPTION = "local (UNIX domain) sockets and named pipes (FIFOs)"

    @classmethod
    def recognizes(cls, file):
        return file.is_socket_or_fifo()

    def get_type(self):
        assert isinstance(self, FilesystemFile)
        st = os.lstat(self.name)
        return stat.S_IFMT(st.st_mode)

    def has_same_content_as(self, other):
        try:
            # (filesystem ID, inode) pair uniquely identifies the socket/pipe
            # Path comparison allows matching against pipes inside an archive
            # (i.e. that would be created by extraction), while using .samefile()
            # lets us match endpoints that might have more than one "canonical"
            # pathname after a mount -o rebind
            if self.get_type() != other.get_type():
                return False
            if os.path.exists(self.name) and os.path.exists(other.name):
                return os.path.samefile(self.name, other.name)
            return os.path.realname(self.name) == os.path.realname(other.name)
        except (AttributeError, OSError):
            # 'other' is likely something odd that doesn't support stat() and/or
            # can't supply an fs_uuid/inode pair for samefile()
            logger.debug(
                "has_same_content: Not a socket, FIFO, or ordinary file: %s",
                other,
            )
            return False

    def create_placeholder(self):
        with get_named_temporary_file(mode="w+", delete=False) as f:
            f.write(format_socket(self.get_type(), self.name))
            f.flush()
            return f.name

    @property
    def path(self):
        if not hasattr(self, "_placeholder"):
            self._placeholder = self.create_placeholder()
        return self._placeholder

    def cleanup(self):
        if hasattr(self, "_placeholder"):
            os.remove(self._placeholder)
            del self._placeholder
        super().cleanup()

    def compare(self, other, source=None):
        with open(self.path) as my_content, open(other.path) as other_content:
            return Difference.from_text_readers(
                my_content,
                other_content,
                self.name,
                other.name,
                source=source,
                comment="socket/FIFO",
            )


def format_socket(mode, filename):
    if stat.S_ISSOCK(mode):
        kind = "UNIX domain socket"
    elif stat.S_ISFIFO(mode):
        kind = "named pipe (FIFO)"
    else:
        kind = "ERROR: problem with an is_socket_or_fifo() predicate"
    return f"{kind}: {filename}\n"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import shutil
import pytest

from diffoscope.comparators.bzip2 import Bzip2File
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

bzip1 = load_fixture("test1.bz2")
bzip2 = load_fixture("test2.bz2")


def test_identification(bzip1):
    assert isinstance(bzip1, Bzip2File)


def test_no_differences(bzip1):
    difference = bzip1.compare(bzip1)
    assert difference is None


@pytest.fixture
def differences(bzip1, bzip2):
    return bzip1.compare(bzip2).details


@skip_unless_tools_exist("bzip2")
def test_content_source(differences):
    assert differences[0].source1 == "test1"
    assert differences[0].source2 == "test2"


@skip_unless_tools_exist("bzip2")
def test_content_source_without_extension(tmpdir, bzip1, bzip2):
    path1 = str(tmpdir.join("test1"))
    path2 = str(tmpdir.join("test2"))
    shutil.copy(bzip1.path, path1)
    shutil.copy(bzip2.path, path2)
    bzip1 = specialize(FilesystemFile(path1))
    bzip2 = specialize(FilesystemFile(path2))
    differences = bzip1.compare(bzip2).details
    assert differences[0].source1 == "test1-content"
    assert differences[0].source2 == "test2-content"


@skip_unless_tools_exist("bzip2")
def test_content_diff(differences):
    expected_diff = get_data("text_ascii_expected_diff")
    assert differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("bzip2")
def test_compare_non_existing(monkeypatch, bzip1):
    assert_non_existing(monkeypatch, bzip1)



#!/usr/bin/env python3

import diffoscope
import json
import sys

from setuptools import setup, find_packages

if sys.version_info < (3, 7):
    print("diffoscope requires at least python 3.7", file=sys.stderr)
    sys.exit(1)


# Load extras_require dict from external JSON file. This allows it to be easily
# shared by the debian/tests/generate-recommends.py script.
with open("extras_require.json") as f:
    extras_require = json.load(f)

setup(
    name="diffoscope",
    version=diffoscope.VERSION,
    description="in-depth comparison of files, archives, and directories",
    long_description=open("README.rst", encoding="utf-8").read(),
    long_description_content_type="text/x-rst",
    author="Diffoscope developers",
    author_email="diffoscope@lists.reproducible-builds.org",
    license="GPL-3+",
    url="https://diffoscope.org/",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"diffoscope": ["scripts/*"]},
    entry_points={
        "console_scripts": ["diffoscope=diffoscope.main:main"],
    },
    install_requires=[
        "python-magic",
        "libarchive-c",
    ],
    extras_require=extras_require,
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Utilities",
    ],
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#project-urls
    project_urls={
        "Issues": "https://salsa.debian.org/reproducible-builds/diffoscope/-/issues",
        "Merge requests": "https://salsa.debian.org/reproducible-builds/diffoscope/-/merge_requests",
    },
)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import get_data


def test_destination(tmpdir):
    def create(x):
        path = os.path.join(str(tmpdir.mkdir(x)), "src")
        os.symlink("/{}".format(x), path)
        return specialize(FilesystemFile(path))

    a = create("a")
    b = create("b")

    expected_diff = get_data("symlink_expected_destination_diff")

    assert a.compare(b).unified_diff == expected_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import shutil
import subprocess

from diffoscope.comparators.gzip import GzipFile
from diffoscope.comparators.sevenz import SevenZFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing

sevenza = load_fixture("test1.disk.gz")
sevenzb = load_fixture("test2.disk.gz")


def sevenz_version():
    try:
        out = subprocess.check_output(["7z"])
    except subprocess.CalledProcessError as e:
        out = e.output
    words = out.decode("UTF-8").split()
    # 7zip 17.04 returns version after "[64]" identifier:
    #   "7-Zip [64] 17.05 : Copyright (c) 1999-2021 Igor Pavlov : 2017-08-28"
    if words[1].startswith("["):
        return words[2].strip()
    return words[1].strip()


def test_identification(sevenza):
    assert isinstance(sevenza, GzipFile)


def test_no_differences(sevenza):
    difference = sevenza.compare(sevenza)
    assert difference is None


@pytest.fixture
def differences(sevenza, sevenzb):
    return sevenza.compare(sevenzb).details


@skip_unless_tools_exist("7z")
def test_content_source(differences):
    # We gzip our test data image, so we need to go one level deeper.
    assert differences[0].source1 == "test1.disk"
    assert differences[0].source2 == "test2.disk"
    assert differences[0].details[0].source1 == "7z l"
    assert differences[0].details[0].source2 == "7z l"


@skip_unless_tools_exist("7z")
def test_content_source_without_extension(tmpdir, sevenza, sevenzb):
    path1 = str(tmpdir.join("test1"))
    path2 = str(tmpdir.join("test2"))
    shutil.copy(sevenza.path, path1)
    shutil.copy(sevenzb.path, path2)
    sevenza = specialize(FilesystemFile(path1))
    sevenzb = specialize(FilesystemFile(path2))
    difference = sevenza.compare(sevenzb).details
    assert difference[0].source1 == "test1-content"
    assert difference[0].source2 == "test2-content"


@skip_unless_tools_exist("7z")
@skip_unless_tool_is_at_least("7z", sevenz_version, "24.05")
def test_metadata_diff(differences):
    assert_diff(differences[0].details[0], "text_sevenzmetadata_expected_diff")


@skip_unless_tools_exist("7z")
def test_compare_non_existing(monkeypatch, sevenza):
    assert_non_existing(monkeypatch, sevenza)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.fontconfig import FontconfigCacheFile

from ..utils.data import load_fixture, assert_diff

cache1 = load_fixture("test1-le64.cache-4")
cache2 = load_fixture("test2-le64.cache-4")


def test_identification(cache1):
    assert isinstance(cache1, FontconfigCacheFile)


def test_no_differences(cache1):
    difference = cache1.compare(cache1)
    assert difference is None


@pytest.fixture
def differences(cache1, cache2):
    return cache1.compare(cache2).details


def test_diff(differences):
    assert_diff(differences[0], "fontconfig_expected_diff")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2024 Jelle van der Waa <jelle@archlinux.org>
# Copyright © 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import re
import subprocess

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.uki import UKIFile
from diffoscope.comparators.utils.command import our_check_output
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import assert_diff, load_fixture
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least

efi_stub = load_fixture("dummyx64.efi.stub")


def ukify_version():
    line = subprocess.check_output(("ukify", "--version")).decode("utf-8")

    return re.search(r"\((.*)\)", line).group(1)


def uki_fixture(prefix, os_release, uname):
    @pytest.fixture
    def uki(tmpdir, efi_stub):
        input_ = str(tmpdir.join(prefix))
        output = str(tmpdir.join("{}.unsigned.efi".format(prefix)))

        with open(input_, "w") as fp:
            fp.write("kernel")

        our_check_output(
            [
                "ukify",
                "build",
                "--stub",
                efi_stub.path,
                "--linux",
                input_,
                "--os-release",
                os_release,
                "--uname",
                uname,
            ],
            cwd=tmpdir,
        )

        return specialize(FilesystemFile(output))

    return uki


uki1 = uki_fixture("linux1", "arch", "6.11")
uki2 = uki_fixture("linux2", "debian", "6.12")


@pytest.fixture
def differences(uki1, uki2):
    return uki1.compare(uki2).details


@skip_unless_tools_exist("objdump")
@skip_unless_tools_exist("ukify")
@skip_unless_tool_is_at_least("ukify", ukify_version, "258~rc3")
def test_no_differences(uki1):
    difference = uki1.compare(uki1)
    assert difference is None


@skip_unless_tools_exist("objdump")
@skip_unless_tools_exist("ukify")
@skip_unless_tool_is_at_least("ukify", ukify_version, "258~rc3")
def test_diff(differences):
    assert_diff(differences[0], "uki_expected_diff")
